/*
   Depth Bias Demo.
   Created by the Programming Ace.
   www.UltimateGameProgramming.com

   The purpose of this demo is to use depth bias to eliminate z fighting with
   polygons that overlap.  We will use a few new D3D functions to do this which
   is in the render function.
*/


// Demo: Depth Bias.


#define WIN32_LEAN_AND_MEAN
#define VC_LEANMEAN

#include<d3dx9.h>


// Function Prototypes...
bool InitializeDirect3D(HWND hwnd, bool fullscreen);// Initialize Direct3D.
bool InitializeObject();                        // Initialize object we are displaying.
void RenderScene();                             // Render the 3D scene.
void SetPerpective();                           // Set the screen perpective.
void ShutdownDirect3D();                        // Shutdown Direct3D.


LPDIRECT3D9 Direct3D_Object = NULL;             // Direct3D object.
LPDIRECT3DDEVICE9 D3D_Device = NULL;            // Device object.
LPDIRECT3DTEXTURE9 Texture = NULL;              // Texture object to hold texture image.

LPDIRECT3DVERTEXBUFFER9 Vertex_Buffer = NULL;   // Vertex buffer for hold our main object.
LPDIRECT3DVERTEXBUFFER9 TexCoord_Buffer = NULL; // Tex coord buffer for our main object.

LPDIRECT3DVERTEXDECLARATION9 VertexDeclaration = NULL;// Vertex Declaration.


// Vertex stucture.
struct Vertex
{
   FLOAT x, y, z;
};

// Texture coordinate structure.
struct TexCoords
{
   FLOAT tu, tv;
};

// Square data for its position and texture coords.
Vertex Polygon_Data[4] =
{
	{-0.3f, 0.5f, 0.0f},
   {0.3f, 0.5f, 0.0f},
	{-0.3f, -0.5f, 0.0f},
   {0.3f, -0.5f, 0.0f}
};

// Data for each points texture coordinate.
TexCoords TexCoords_Data[4] =
{
   {0.0, 0.0},
   {1.0, 0.0},
	{0.0, 1.0},
   {1.0, 1.0}
};

// Mouse rotations.
float xRotation = 0.0f;
float yRotation = 0.0f;

// Values for the depth bias.
float zScale = -0.1f;
float biasAmount = 0.0f;

// Used to convert a float to a DWORD.
inline DWORD F2DW(float f){ return *((DWORD*)&f); }


LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
   static POINT oldMousePos;     // Last mouse position.
   static POINT currentMousePos; // Current mouse position.
   static bool  isMouseActive;   // Is the left mouse button down.

   switch(message)
      {
		   case WM_KEYDOWN:
		      switch(wParam)
               { 
                  // Close the app if the user presses escape.
		            case VK_ESCAPE:
                     PostQuitMessage(0);
				         break;

                  case VK_UP:
                     zScale += 0.1f;
                     break;

                  case VK_DOWN:
                     zScale -= 0.1f;
                     break;

                  case VK_LEFT:
                     biasAmount += 0.1f;
                     break;

                  case VK_RIGHT:
                     biasAmount -= 0.1f;
                     break;
               }
            break;


         case WM_DESTROY:
         case WM_CLOSE:
            PostQuitMessage(0);
            break;


         case WM_LBUTTONDOWN:
				oldMousePos.x = currentMousePos.x = LOWORD (lParam);
				oldMousePos.y = currentMousePos.y = HIWORD (lParam);
            isMouseActive = true;
			   break;


			case WM_LBUTTONUP:
				isMouseActive = false;
				break;


			case WM_MOUSEMOVE:
				currentMousePos.x = LOWORD (lParam);
				currentMousePos.y = HIWORD (lParam);

				if(isMouseActive)
					{
						xRotation -= (currentMousePos.x - oldMousePos.x);
						yRotation -= (currentMousePos.y - oldMousePos.y);
					}

				oldMousePos.x = currentMousePos.x;
				oldMousePos.y = currentMousePos.y;
			   break;


            default:
               break;
       }

    return DefWindowProc( hwnd, message, wParam, lParam );
}


int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nShowCmd)
{
   MSG msg;                                           // Message object.
   HWND hwnd;                                         // Handle to the window.
   WNDCLASSEX windowClass;                            // Window class object.
   bool done = false;                                 // true if we should exit.

   // This is the Window class.
   windowClass.cbSize = sizeof(WNDCLASSEX);           // size of the WNDCLASSEX structure.
   windowClass.style = CS_HREDRAW | CS_VREDRAW;       // style of the window.
   windowClass.lpfnWndProc = WndProc;                 // Address to the windows procedure.
   windowClass.cbClsExtra = 0;                        // Extra class information.
   windowClass.cbWndExtra = 0;                        // Extra window information.
   windowClass.hInstance = hInstance;                 // Handle of application Instance.
   windowClass.hIcon = LoadIcon(NULL, IDI_APPLICATION);// Window Icon.
   windowClass.hCursor = LoadCursor(NULL, IDC_ARROW); // Load mouse cursor.
   windowClass.hbrBackground = NULL;                  // Background color.
   windowClass.lpszMenuName = NULL;                   // Menu.
   windowClass.lpszClassName = "UGPClass";            // Name of the window class.
   windowClass.hIconSm = LoadIcon(NULL, IDI_APPLICATION);// Minimized window icon.

   // You must register you class with Windows.
   if(!RegisterClassEx(&windowClass)) return 0;

   // After your class has been registered then you are ready to create your window.
   hwnd = CreateWindowEx(NULL,                                       // The extended style.
                         "UGPClass",                                 // Window class.
                         "Depth Bias - by The Programming Ace",      // Window name.
                          WS_OVERLAPPEDWINDOW | WS_VISIBLE |         // Window style.
                          WS_SYSMENU |WS_CLIPCHILDREN |              // Window style.
                          WS_CLIPSIBLINGS,                           // Window style.
                          100, 100,                                  // X, Y coords.
                          640, 480,                                  // Window size.
                          NULL,                                      // Handle to parent window.
                          NULL,                                      // Menu.
                          hInstance,                                 // Handle to app instance.
                          NULL);                                     // Pointer to window.

   // If there was an error with creating the window, then close the program.
   if(!hwnd) return 0;

   ShowWindow(hwnd, SW_SHOW);    // Show the window.
   UpdateWindow(hwnd);           // Update its display.

   done = false;                 // false = run program, true means stop.

   // Initialize Direct3D.  If fails then we don't want to run the application.
   if(!InitializeDirect3D(hwnd, false)) done = true;

   // Application loop.
   while(!done)
      {
         if(PeekMessage(&msg, NULL, NULL, NULL, PM_REMOVE))
            {
               // If a quit message is received then stop rendering and quit the app.
               if(msg.message == WM_QUIT)
                  {
                     done = true;
                  }

               TranslateMessage(&msg);
               DispatchMessage(&msg);
            }
         else
            {
               RenderScene();
            }
      }

    // Here we shutdown Direct 3D.
	ShutdownDirect3D();

   // Here we unregister the window class with the OS.
   UnregisterClass("UGPClass", windowClass.hInstance);

   return (int)msg.wParam;
}


bool InitializeDirect3D(HWND hwnd, bool fullscreen)
{
   // This object will allow us to set the display mode of the screen.
   D3DDISPLAYMODE DisplayMode;

   // This will allow us to set the parameters of the screen.
   D3DPRESENT_PARAMETERS Present_Parameters;

   // This is used to get the capabilities of the hardware.
   D3DCAPS9 D3DCaps;

   // It is always a good idea to clear out memory in object although not necessary.
   ZeroMemory(&Present_Parameters, sizeof(Present_Parameters));

   // Create the Direct3D object to get everything started.
   Direct3D_Object = Direct3DCreate9(D3D_SDK_VERSION);

   // Error checking.  Make sure that it was successful.
   if(Direct3D_Object == NULL)
      {
         MessageBox(NULL, "Error, couldn't initialize DirectX!?!",
                    "Error!", MB_OK);
         return false;
      }

   // This function will get the display mode of the device and place it in DisplayMode.
   if(FAILED(Direct3D_Object->GetAdapterDisplayMode(D3DADAPTER_DEFAULT, &DisplayMode)))
      {
         MessageBox(NULL, "Error setting the display mode.", "Error!", MB_OK);
         return false;
      }

   // Get the capabilities of the hardware.
	if(FAILED(Direct3D_Object->GetDeviceCaps(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, &D3DCaps)))
		{
		   return false;
	   }

   // Test which is supported, hardware or software vertex processing.
	DWORD VertexProcessing = 0;

	if(D3DCaps.VertexProcessingCaps != 0)
		VertexProcessing |= D3DCREATE_HARDWARE_VERTEXPROCESSING;
	else
		VertexProcessing |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;

   // Here we are setting the applications parameters...
   if(fullscreen)
      {
         Present_Parameters.Windowed = FALSE;               // Window mode (fullscreen).
         Present_Parameters.BackBufferWidth = DisplayMode.Width;
         Present_Parameters.BackBufferHeight = DisplayMode.Height;
      }
   else
      Present_Parameters.Windowed = TRUE;                   // Window mode (not fullscreen).
   Present_Parameters.SwapEffect = D3DSWAPEFFECT_DISCARD;   // Dealing with animation (see doc).
   Present_Parameters.BackBufferFormat = DisplayMode.Format;// Render to the area of the screen.
   Present_Parameters.BackBufferCount = 1;                  // Number of back buffers.
   Present_Parameters.EnableAutoDepthStencil = TRUE;        // Check documentation.
   Present_Parameters.AutoDepthStencilFormat = D3DFMT_D16;  // Check documentation.

   // Now we must create the rendering device.
   if(FAILED(Direct3D_Object->CreateDevice(D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hwnd,
                  VertexProcessing, &Present_Parameters, &D3D_Device)))
      {
         MessageBox(NULL, "CreateDevice() failed!  Make sure you have DirectX 9.",
                    "Error!", MB_OK);
         return false;
      }

   D3D_Device->SetRenderState(D3DRS_LIGHTING, FALSE);       // Lighting off.
	D3D_Device->SetRenderState(D3DRS_CULLMODE, D3DCULL_NONE);// Culling off.
   D3D_Device->SetRenderState(D3DRS_ZENABLE, TRUE);         // Enable depth testing.

   // Here we set up the object this demo will render (for example a polygon).
   if(!InitializeObject()) return false;

   // Set the image states to get a good quality image.
	D3D_Device->SetSamplerState(0, D3DSAMP_MINFILTER, D3DTEXF_LINEAR);
	D3D_Device->SetSamplerState(0, D3DSAMP_MAGFILTER, D3DTEXF_LINEAR);

   // Set the perpective.
   SetPerpective();

   return true;
}


bool InitializeObject()
{
   // Create the vertex buffer that will hold the square.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(Polygon_Data), 0, 0,
                           D3DPOOL_DEFAULT, &Vertex_Buffer, NULL)))
      return false;
	
   // Pointer to the buffer.
   Vertex* Vertices;

   // Lock the buffer we can write to it.
   if(FAILED(Vertex_Buffer->Lock(0, sizeof(Polygon_Data), (void**)&Vertices, 0)))
      return false;

   // Here we copy the square's data into the vertex buffer.
   memcpy(Vertices, Polygon_Data, sizeof(Polygon_Data));

   // Unlock when your done coping data into the buffer.
   Vertex_Buffer->Unlock();


   // Create the vertex buffer that will hold the texture coordinates.
	if(FAILED(D3D_Device->CreateVertexBuffer(sizeof(TexCoords_Data), 0, 0,
                           D3DPOOL_DEFAULT, &TexCoord_Buffer, NULL)))
      return false;
	
   // Pointer to the buffer.
   TexCoords* TexC;

   // Lock the buffer we can write to it.
   if(FAILED(TexCoord_Buffer->Lock(0, sizeof(TexCoords_Data), (void**)&TexC, 0)))
      return false;

   // Copy the tex coords into the buffer for the tex coords.
   memcpy(TexC, TexCoords_Data, sizeof(TexCoords_Data));

   // Unlock when your done coping data into the buffer.
   TexCoord_Buffer->Unlock();

   // Here we describe to Direct3D how a vertex is to be described.  This makes up the stream
   // number, offset, type, method, usage, and the usage index (consult the documentation if you
   // wish to know more).
   D3DVERTEXELEMENT9 Declaration[] = 
	   {
	      {0, 0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0},
		   {1, 0, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0},
		   D3DDECL_END()
	   };

   // Create the vertex declaration so DirectX knows how our vertex is made up.
   D3D_Device->CreateVertexDeclaration(Declaration, &VertexDeclaration);


   // Here we load the texture into the texture object.
   if(D3DXCreateTextureFromFile(D3D_Device, "ugp.tga", &Texture) != D3D_OK)
      return false;

   return true;
}


void SetPerpective()
{
   // This matrix will do nothing more but tell Direct3D how far back it can see
   // and how close it can see.
   D3DXMATRIXA16 Projection;

   // Set the perspective information.
   D3DXMatrixPerspectiveFovLH(&Projection, 45.0f, 640/480, 0.1f, 500.0f);

   // Set the projection.
   D3D_Device->SetTransform(D3DTS_PROJECTION, &Projection);
}


void RenderScene()
{
   // Clear the screen to black.
   D3D_Device->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                     D3DCOLOR_XRGB(0,0,0), 1.0f, 0);
    
   // Tell Direct 3D we are now going to start drawing.
   D3D_Device->BeginScene();

      D3DXMATRIX matTrans, matRot, matWorld, matScale;

      // Set the rotation by mouse.
      D3DXMatrixTranslation(&matTrans, 0.0f, 0.0f, 2.0f);
      D3DXMatrixRotationYawPitchRoll(&matRot, D3DXToRadian(xRotation),
                                     D3DXToRadian(yRotation), 0.0f);
      matWorld = matRot * matTrans;

      // Apply translation and rotation.
      D3D_Device->SetTransform(D3DTS_WORLD, &matWorld);

      // Set the texture image.
      D3D_Device->SetTexture(0, Texture);

      // This will bind the vertex data in the buffer to the Direct3D device.
      D3D_Device->SetStreamSource(0, Vertex_Buffer, 0, sizeof(Vertex));
      D3D_Device->SetStreamSource(1, TexCoord_Buffer, 0, sizeof(TexCoords));

      // Set the vertex stream declaration.
      D3D_Device->SetVertexDeclaration(VertexDeclaration);

      // This will draw everything in the buffer.
      D3D_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);


      // Next will draw the object again at the same location but this time with depth bias.
      D3D_Device->SetRenderState(D3DRS_SLOPESCALEDEPTHBIAS, F2DW(zScale));
      D3D_Device->SetRenderState(D3DRS_DEPTHBIAS, F2DW(biasAmount));

      D3DXMatrixScaling(&matScale, 0.5f, 0.5f, 0.0f);

      matWorld = matScale * matWorld;
      D3D_Device->SetTransform(D3DTS_WORLD, &matWorld);

      // We will no be displaying a texture.
      D3D_Device->SetTexture(0, NULL);

      // This will draw everything in the buffer.
      D3D_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);

      // Before we finish up lets turn off depth bias.
      D3D_Device->SetRenderState(D3DRS_SLOPESCALEDEPTHBIAS, F2DW(0.0f));
      D3D_Device->SetRenderState(D3DRS_DEPTHBIAS, F2DW(0.0f));
    
   // We are done drawing for this scene.
   D3D_Device->EndScene();
    
   // Swap the old frame with the new one.
   D3D_Device->Present(NULL, NULL, NULL, NULL);
}


void ShutdownDirect3D()
{
   // Here we release the Direct3D objects.
   if(Texture != NULL) 
      {
         Texture->Release();
         Texture = NULL;
      }

   if(D3D_Device != NULL)
      {
         D3D_Device->Release();
         D3D_Device = NULL;
      }

   if(Direct3D_Object != NULL)
      {
         Direct3D_Object->Release();
         Direct3D_Object = NULL;
      }

   if(Vertex_Buffer != NULL)
      {
	      Vertex_Buffer->Release();
         Vertex_Buffer = NULL;
      }

   if(TexCoord_Buffer != NULL)
      {
	      TexCoord_Buffer->Release();
         TexCoord_Buffer = NULL;
      }

   if(VertexDeclaration != NULL)
      {
         VertexDeclaration->Release();
         VertexDeclaration = NULL;
      }
}


// Recap:

// When running this application without the depth bias you can see a hugh difference
// in the scene.  Without depth bias things look really bad really fast.


// Copyright September 2004
// All Rights Reserved!
// Allen Sherrod
// ProgrammingAce@UltimateGameProgramming.com
// www.UltimateGameProgramming.com